# -*- coding: utf-8 -*-
"""
Created on Tue Jun 30 21:14:00 2020

@author: yrc2
"""
from biorefineries import LAOs

__all__ = (
    'get_product_purity',
    'get_product_flow',
    'get_alcohol_loss',
    'get_LAOs_MPSP',
    'set_LAOs_MPSP',
)

fatty_alcohols = ('Hexanol','Octanol', 'Decanol', 'Dodecanol', 'Tetradecanol')
products = ('Hexene','Octene', 'Decene', 'Dodecene', 'Tetradecene')

def get_product_purity(product):
    """Get product purity in olefin wt. %."""
    return product.imass[products].sum() / product.F_mass
        
def get_product_flow():
    """Get product flow in ton / yr."""
    return sum([i.get_total_flow('ton/day') for i in LAOs.products]) * LAOs.LAOs_tea.operating_days
    
def get_alcohol_loss():
    """Get alcohol loss in ton / yr."""
    F_ton_per_yr = LAOs.wastewater.get_flow('ton/day', fatty_alcohols)
    return F_ton_per_yr * LAOs.LAOs_tea.operating_days

def get_LAOs_MPSP():
    """Get product price in USD/ton."""
    products = LAOs.products
    sales = sum([i.cost for i in products])
    F_mass_LAOs = sum([i.F_mass for i in products])
    LAOs_tea = LAOs.LAOs_tea
    return (LAOs_tea.solve_sales() / LAOs_tea.operating_days / 24  + sales) / F_mass_LAOs * 907.185 # To USD / ton

def set_LAOs_MPSP(MPSP):
    """Set product price in USD/ton."""
    price = MPSP / 907.185
    for i in LAOs.products:
        i.price = price


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

# =============================================================================
# Setup
# =============================================================================

import numpy as np
import biosteam as bst
from biosteam.evaluation import Model, Metric
from chaospy import distributions as shape
from ._settings import CFs
from ._utils import set_yield, _feedstock_factor
from .systems import simulate_and_print, \
    SSCF_flowsheet, SSCF_groups, SSCF_teas, SSCF_funcs, \
    SHF_flowsheet, SHF_groups, SHF_teas, SHF_funcs

__all__ = ('create_model',)


# %%

# =============================================================================
# Models for uncertainty and sensitivity analyses
# =============================================================================

def create_model(kind='SSCF'):
    if 'SSCF' in str(kind).upper():
        flowsheet = SSCF_flowsheet
        groups = SSCF_groups
        teas = SSCF_teas
        funcs = SSCF_funcs
    elif 'SHF' in str(kind).upper():
        flowsheet = SHF_flowsheet
        groups = SHF_groups
        teas = SHF_teas
        funcs = SHF_funcs
    else:
        raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')

    bst.main_flowsheet.set_flowsheet(flowsheet)
    s = flowsheet.stream
    u = flowsheet.unit
    sys = flowsheet.system

    simulate_and_print(kind)
    # =============================================================================
    # Overall biorefinery metrics
    # =============================================================================

    # Minimum product selling price of lactic_acid stream
    lactic_acid = s.lactic_acid
    lactic_tea = teas['lactic_tea']
    def get_MPSP():
        lactic_acid.price = 0
        for i in range(3):
            MPSP = lactic_acid.price = lactic_tea.solve_price(lactic_acid)
        return MPSP

    feedstock = s.feedstock
    # Yield in 10^6 kg/yr
    # lactic_no_CHP_tea = teas['lactic_no_CHP_tea']
    get_annual_factor = lambda: lactic_tea.operating_days*24
    get_total_yield = lambda: lactic_acid.F_mass*get_annual_factor()/1e6
    # Yield in % of dry feedstock
    get_mass_yield = lambda: lactic_acid.F_mass/(feedstock.F_mass-feedstock.imass['Water'])
    R301 = u.R301
    get_titer = lambda: R301.effluent_titer
    # Purity (%) of LacticAcid in the final product
    get_purity = lambda: lactic_acid.imass['LacticAcid']/lactic_acid.F_mass
    # Recovery (%) = recovered/amount in fermentation broth
    get_recovery = lambda: lactic_acid.imol['LacticAcid'] \
        /(R301.outs[0].imol['LacticAcid']+2*R301.outs[0].imol['CalciumLactate'])
    get_overall_TCI = lambda: lactic_tea.TCI/1e6
    get_lactic_sale = lambda: get_total_yield()*lactic_acid.price
    # Including negative product sales (ash/gypsum disposal) but excluding electricity credit
    gypsum = s.gypsum
    get_gypsum_sale = lambda: gypsum.F_mass*gypsum.price*get_annual_factor()/1e6
    ash = s.ash
    get_ash_sale = lambda: ash.F_mass*ash.price*get_annual_factor()/1e6
    get_operating_cost = lambda: lactic_tea.AOC/1e6-get_gypsum_sale()-get_ash_sale()
    # Including negative product sales (ash/gypsum disposal) but excluding electricity credit
    get_material_cost = lambda: lactic_tea.material_cost/1e6-get_gypsum_sale()-get_ash_sale()
    # System power usage, individual unit power usage should be positive
    CHP = u.CHP
    excess_power = lambda: CHP.electricity_generated
    electricity_price = bst.PowerUtility.price
    # Electricity credit is positive if getting revenue from excess electricity
    get_electricity_credit = lambda: (excess_power()*electricity_price*get_annual_factor())/1e6

    metrics = [Metric('MPSP', get_MPSP, '$/kg'),
               Metric('Total product yield', get_total_yield, '10^6 kg/yr'),
               Metric('Product mass yield', get_mass_yield, '%'),
               Metric('Fermentation titer', get_titer, 'g/L'),
               Metric('Product purity', get_purity, '%'),
               Metric('Product recovery', get_recovery, '%'),
               Metric('Total capital investment', get_overall_TCI, '10^6 $'),
               Metric('Annual operating cost', get_operating_cost, '10^6 $/yr'),
               Metric('Annual material cost', get_material_cost, '10^6 $/yr'),
               Metric('Annual product sale', get_lactic_sale, '10^6 $/yr'),
               Metric('Annual electricity credit', get_electricity_credit, '10^6 $/yr')
               ]

    ####################### Capital cost breakdown #######################
    def get_installed_cost(group):
        return lambda: group.get_installed_cost()
    for group in groups:
        if group.name == 'preprocessing_group': continue
        metrics.append(
            Metric(group.name, get_installed_cost(group), '10^6 $', 'Installed cost'))

    # All checks should be ~0
    check_installed_cost = \
        lambda: sum(get_installed_cost(group)()
                    for group in groups) - lactic_tea.installed_equipment_cost/1e6
    metrics.append(Metric('Check', check_installed_cost, '10^6 $', 'Installed cost'))

    ############### Material cost and product sale breakdown ###############

    TEA_feeds = teas['TEA_feeds']
    TEA_products = teas['TEA_products']

    def get_material_cost(feed):
        return lambda: feed.price*feed.F_mass*get_annual_factor()/1e6
    for feed in TEA_feeds:
        metrics.append(Metric(feed.ID, get_material_cost(feed), '10^6 $/yr', 'Material cost'))

    # Sulfuric acid used in pretreatment and separation processes
    T602_S = u.T602_S
    get_pretreatment_sulfuric_acid_ratio = lambda: T602_S.outs[0].F_mass/T602_S.F_mass_out
    # Ammonia used in pretreatment and CHP process
    T603_S = u.T603_S
    get_pretreatment_ammonia_ratio = lambda: T603_S.outs[0].F_mass/T602_S.F_mass_out

    check_material_cost = lambda: sum(get_material_cost(feed)()
                                      for feed in TEA_feeds) - lactic_tea.material_cost/1e6

    metrics.extend((
        Metric('Pretreatment sulfuric acid ratio', get_pretreatment_sulfuric_acid_ratio,
               '%', 'Material cost'),
        Metric('Pretreatment ammonia ratio', get_pretreatment_ammonia_ratio,
               '%', 'Material cost'),
        Metric('Check', check_material_cost, '10^6 $/yr', 'Material cost')
        ))

    def get_product_sale(stream):
        return lambda: stream.price*stream.F_mass*get_annual_factor()/1e6
    for product in TEA_products:
        metrics.append(Metric(product.ID, get_product_sale(product), '10^6 $/yr', 'Product sale'))
    check_product_sale= \
        lambda: sum(get_product_sale(product)() for product in TEA_products) \
            - lactic_tea.sales/1e6
    metrics.append(Metric('Check', check_product_sale, '10^6 $/yr', 'Product sale'))


    ########## Heating demand breakdown (positive if needs heating) ##########
    get_system_heating_demand = lambda: CHP.system_heating_demand/1e6
    get_pretreatment_steam_heating_demand = lambda: CHP.side_streams_lps.duty/1e6
    HXN = u.HXN
    get_HXN_heating_demand = lambda: sum(i.duty for i in HXN.heat_utilities
                                        if i.flow*i.duty>0)/1e6
    get_CHP_heating_demand = lambda: sum(i.duty for i in CHP.heat_utilities
                                        if i.flow*i.duty>0)/1e6

    def get_heating_demand(group):
        return lambda: group.get_heating_duty()

    for group in groups:
        if group.name in ('preprocessing_group', 'HXN_group', 'CHP_group'): continue
        # The only heating demand for the pretreatment system is the heat needed to
        # generate the side steam
        if group.name == 'pretreatment_group':
            metrics.append(Metric(group.name, get_pretreatment_steam_heating_demand, '10^6 kJ/hr',
                                    'Heating demand'))
        else: metrics.append(Metric(group.name, get_heating_demand(group), '10^6 kJ/hr',
                                      'Heating demand'))

    check_heating_demand = \
        lambda: sum((get_heating_demand(group)() for group in groups),
                    get_pretreatment_steam_heating_demand()) + \
                    get_HXN_heating_demand() + get_CHP_heating_demand()

    metrics.extend((
        Metric('HXN', get_HXN_heating_demand, '10^6 kJ/hr', 'Heating demand'),
        Metric('CHP', get_CHP_heating_demand, '10^6 kJ/hr', 'Heating demand'),
        Metric('Sum', get_system_heating_demand, '10^6 kJ/hr', 'Heating demand'),
        Metric('Check', check_heating_demand, '10^6 kJ/hr', 'Heating demand')
        ))

    ########## Cooling demand breakdown (negative if needs cooling) ##########
    CT = u.CT
    get_system_cooling_water_duty = lambda: CT.system_cooling_water_duty/1e6
    get_HXN_cooling_demand = lambda: sum(i.duty for i in HXN.heat_utilities
                                        if i.flow*i.duty<0)/1e6
    get_CT_cooling_water_duty = lambda: sum(i.duty for i in CT.heat_utilities
                                        if i.flow*i.duty<0)/1e6

    def get_cooling_demand(group):
        return lambda: -group.get_cooling_duty()

    for group in groups:
        if group.name in ('preprocessing_group', 'HXN_group', 'CT_group'): continue
        else: metrics.append(Metric(group.name, get_cooling_demand(group),
                                      '10^6 kJ/yr', 'Cooling demand'))

    check_cooling_demand = \
        lambda: sum(get_cooling_demand(group)() for group in groups) + \
                    get_HXN_cooling_demand() + get_CT_cooling_water_duty()

    metrics.extend((
        Metric('HXN', get_HXN_cooling_demand, '10^6 kJ/hr', 'Cooling demand'),
        Metric('CT', get_CT_cooling_water_duty, '10^6 kJ/hr', 'Cooling demand'),
        Metric('Sum', get_system_cooling_water_duty, '10^6 kJ/hr', 'Cooling demand'),
        Metric('Check', check_cooling_demand, '10^6 kJ/hr', 'Cooling demand')
        ))


    ########### Power demand breakdown (positive if using power) ###########
    lactic_sys = sys.lactic_sys
    get_electricity_use = funcs['get_electricity_use']

    def get_power_demand(group):
        return lambda: sum(i.rate for i in group.power_utilities)

    for group in groups:
        if group.name == 'preprocessing_group': continue
        metrics.append(Metric(group.name, get_power_demand(group), 'kW', 'Power demand'))

    check_power_demand = lambda: sum(get_power_demand(group)()
                                     for group in groups) - get_electricity_use()
    metrics.extend((
        Metric('Sum', get_electricity_use, 'kW', 'Power demand'),
        Metric('Check', check_power_demand, 'kW', 'Power demand')
        ))

    ###### Utility cost breakdown (including heating, cooling, and power) ######
    get_system_utility_cost = lambda: lactic_tea.utility_cost/1e6

    def get_utility_cost(group):
        return lambda: sum(i.utility_cost for i in group.units)*get_annual_factor()/1e6

    for group in groups:
        if group.name == 'preprocessing_group': continue
        metrics.append(Metric(group.name, get_utility_cost(group), '10^6 $/yr', 'Utility cost'))

    check_utility_cost = \
        lambda: sum(get_utility_cost(group)() for group in groups) \
            - get_system_utility_cost()

    metrics.extend((
        Metric('Sum', get_system_utility_cost, '10^6 $/yr', 'Utility cost'),
        Metric('Check', check_utility_cost, '10^6 $/yr', 'Utility cost')
        ))

    # To see if TEA converges well for each simulation
    get_NPV = lambda: lactic_tea.NPV
    metrics.append(Metric('NPV', get_NPV, '$', 'TEA'))

    model_dct = {'index_TEA': len(metrics)}

    ######## Evalute system at different internal rates of return ########
    def create_IRR_metrics(IRR):
        def get_IRR_based_MPSP():
            lactic_tea.IRR = IRR
            return get_MPSP()
        return [Metric('MPSP', get_IRR_based_MPSP, '$/kg', f'IRR={IRR:.0%}'),
                Metric('NPV', get_NPV, '$', f'IRR={IRR:.0%}')]

    # This is to ensure Monte Carlo results will be at 10% IRR
    IRR1 = np.arange(0, 0.10, 0.01)
    IRR2 = np.arange(0.11, 0.41, 0.01)
    model_dct['IRRs'] = IRRs = IRR1.tolist() + IRR2.tolist() + [0.1]
    for IRR in IRRs:
        metrics.extend((i for i in create_IRR_metrics(IRR)))

    model_dct['index_IRR'] = len(metrics)

    ######################## Global warming potential ########################
    get_GWP = funcs['get_GWP']
    get_FEC = funcs['get_FEC']
    metrics.extend((
        Metric('Total GWP', get_GWP, 'kg CO2-eq/kg', 'LCA'),
        Metric('Total FEC', get_FEC, 'MJ/kg', 'LCA')
        ))

    get_material_GWP = funcs['get_material_GWP']
    get_electricity_GWP = funcs['get_electricity_GWP']

    natural_gas = s.natural_gas
    get_CH4_production_GWP = lambda: \
        CFs['GWP_CFs']['CH4']*natural_gas.F_mass/lactic_acid.F_mass
    get_CH4_onsite_GWP = lambda: \
         natural_gas.get_atomic_flow('C')*natural_gas.chemicals.CO2.MW/lactic_acid.F_mass

    lime = s.lime
    lime_CHP = s.lime_CHP
    get_lime_GWP  = lambda: \
        CFs['GWP_CFs']['Lime']*(lime.F_mass+lime_CHP.F_mass)/lactic_acid.F_mass

    ethanol = s.ethanol
    get_ethanol_onsite_GWP = lambda: \
         ethanol.get_atomic_flow('C')*ethanol.chemicals.CO2.MW/lactic_acid.F_mass

    # For all other materials, including both production and combustion
    get_other_materials_GWP = lambda: get_material_GWP()+get_ethanol_onsite_GWP()-\
        get_CH4_production_GWP()-get_lime_GWP()

    check_GWP = lambda: get_GWP()-get_material_GWP()-get_electricity_GWP()- \
        get_CH4_onsite_GWP()-get_ethanol_onsite_GWP()

    metrics.extend((
        Metric('Electricity', get_electricity_GWP, 'kg CO2-eq/kg', 'GWP'),
        Metric('Natural gas production', get_CH4_production_GWP, 'kg CO2-eq/kg', 'GWP'),
        Metric('Natural gas combustion', get_CH4_onsite_GWP, 'kg CO2-eq/kg', 'GWP'),
        Metric('Lime', get_lime_GWP, 'kg CO2-eq/kg', 'GWP'),
        Metric('Other materials', get_other_materials_GWP, 'kg CO2-eq/kg', 'GWP'),
        Metric('GWP check', check_GWP, 'kg CO2-eq/kg', 'GWP')
        ))

    ######################## Fossil energy consumption ########################
    get_electricity_FEC = funcs['get_electricity_FEC']
    get_CH4_FEC = lambda: CFs['FEC_CFs']['CH4']*natural_gas.F_mass/lactic_acid.F_mass
    get_lime_FEC = lambda: \
        CFs['FEC_CFs']['Lime']*(lime.F_mass+lime_CHP.F_mass)/lactic_acid.F_mass

    get_material_FEC = funcs['get_material_FEC']
    get_other_materials_FEC = lambda: get_material_FEC()-get_CH4_FEC()-get_lime_FEC()

    check_FEC = lambda: get_FEC() - get_material_FEC() - \
        get_system_utility_cost()*1e6/get_annual_factor()/bst.PowerUtility.price * \
        CFs['FEC_CFs']['Electricity']/lactic_acid.F_mass

    metrics.extend((
        Metric('Electricity', get_electricity_FEC, 'MJ/kg', 'FEC'),
        Metric('Natural gas', get_CH4_FEC, 'MJ/kg', 'FEC'),
        Metric('Lime', get_lime_FEC, 'MJ/kg', 'FEC'),
        Metric('Other materials', get_other_materials_FEC, 'MJ/kg', 'FEC'),
        Metric('FEC check', check_FEC, 'MJ/kg', 'FEC')
        ))

    # =============================================================================
    # Construct base model and add parameters
    # =============================================================================

    model_dct['model'] = model = Model(lactic_sys, metrics)

    param = model.parameter

    def baseline_uniform(baseline, ratio):
        return shape.Uniform(baseline*(1-ratio), baseline*(1+ratio))

    def baseline_triangle(baseline, ratio):
        return shape.Triangle(baseline*(1-ratio), baseline, baseline*(1+ratio))

    # A fake parameter serving as a "blank" in sensitivity analysis to capture
    # fluctuations due to converging errors
    D = baseline_uniform(1, 0.1)
    @param(name='Blank parameter', element=feedstock, kind='coupled', units='',
           baseline=1, distribution=D)
    def set_blank_parameter(anything):
        # This does nothing
        feedstock.T = feedstock.T

    ######################## TEA parameters ########################
    # U101 = SSCF.U101
    # D = baseline_uniform(2205, 0.1)
    # @param(name='Feedstock flow rate', element=feedstock, kind='coupled', units='dry-ton/day',
    #        baseline=2205, distribution=D)
    # def set_feedstock_flow_rate(rate):
    #     feedstock.mass *= rate / U101._cached_flow_rate
    #     U101._cached_flow_rate = rate

    D = shape.Triangle(0.84, 0.9, 0.96)
    @param(name='Plant uptime', element='TEA', kind='isolated', units='%',
           baseline=0.9, distribution=D)
    def set_plant_uptime(uptime):
        lactic_tea.operating_days = 365 * uptime

    D = baseline_triangle(1, 0.25)
    @param(name='TCI ratio', element='TEA', kind='isolated', units='% of baseline',
            baseline=1, distribution=D)
    def set_TCI_ratio(new_ratio):
        old_ratio = lactic_tea._TCI_ratio_cached
        # old_ratio = lactic_no_CHP_tea._TCI_ratio_cached
        for unit in lactic_sys.units:
            if hasattr(unit, 'cost_items'):
                for item in unit.cost_items:
                    unit.cost_items[item].cost /= old_ratio
                    unit.cost_items[item].cost *= new_ratio
        lactic_tea._TCI_ratio_cached = new_ratio
        # lactic_no_CHP_tea._TCI_ratio_cached = new_ratio

    # Only include materials that account for >5% of total annual material cost,
    # enzyme not included as it's cost is more affected by the loading (considered later)
    D = shape.Triangle(60, 71.3, 83.7)
    @param(name='Feedstock unit price', element='TEA', kind='isolated', units='$/dry-ton',
           baseline=71.3, distribution=D)
    def set_feedstock_price(price):
        feedstock.price = price / _feedstock_factor

    sulfuric_acid = s.sulfuric_acid
    D = shape.Triangle(0.0910, 0.0948, 0.1046)
    @param(name='Sulfuric acid unit price', element='TEA', kind='isolated', units='$/kg',
           baseline=0.0948, distribution=D)
    def set_sulfuric_acid_price(price):
        sulfuric_acid.price = price

    D = shape.Triangle(0.160, 0.262, 0.288)
    @param(name='Lime unit price', element='TEA', kind='isolated', units='$/kg',
           baseline=0.262, distribution=D)
    def set_lime_price(price):
        lime.price = price
        lime_CHP.price = price

    D = shape.Triangle(0.198, 0.253, 0.304)
    @param(name='Natural gas unit price', element='TEA', kind='isolated', units='$/kg',
           baseline=0.253, distribution=D)
    def set_natural_gas_price(price):
        natural_gas.price = price

    D = shape.Uniform(-0.0288, 0.00776)
    @param(name='Gypsum unit price', element='TEA', kind='isolated', units='$/kg',
           baseline=0, distribution=D)
    def set_gypsum_price(price):
        gypsum.price = price

    D = shape.Triangle(0.067, 0.070, 0.074)
    @param(name='Electricity unit price', element='TEA', kind='isolated', units='$/kWh',
           baseline=0.070, distribution=D)
    def set_electricity_price(price):
        bst.PowerUtility.price = price

    ######################## Pretreatment parameters ########################
    M202 = u.M202
    D = shape.Triangle(0.25, 0.3, 0.4)
    @param(name='Pretreatment solid loading', element=M202, kind='coupled', units='%',
           baseline=0.3, distribution=D)
    def set_pretreatment_solid_loading(loading):
        M202.solid_loading = loading

    T201 = u.T201
    D = shape.Triangle(10, 22.1, 35)
    @param(name='Pretreatment sulfuric acid loading', element=T201,
           kind='coupled', units='mg/g', baseline=22.1, distribution=D)
    def set_pretreatment_sulfuric_acid_loading(loading):
        T201.feedstock_dry_mass = feedstock.F_mass - feedstock.imass['H2O']
        T201.acid_loading = loading

    R201 = u.R201
    D = shape.Triangle(0.06, 0.099, 0.12)
    @param(name='Pretreatment glucan-to-glucose', element=R201, kind='coupled', units='%',
           baseline=0.099, distribution=D)
    def set_R201_glucan_conversion(X):
        R201.pretreatment_rxns[0].X = X

    D = shape.Triangle(0.8, 0.9, 0.92)
    @param(name='Pretreatment xylan-to-xylose', element=R201, kind='coupled', units='%',
           baseline=0.9, distribution=D)
    def set_R201_xylan_conversion(X):
        R201.pretreatment_rxns[4].X = X

    ######################## Conversion parameters ########################
    M301 = u.M301
    D = shape.Triangle(0.175, 0.2, 0.25)
    @param(name='Enzymatic hydrolysis solid loading', element=M301, kind='coupled', units='%',
           baseline=0.2, distribution=D)
    def set_R301_solid_loading(loading):
        M301.solid_loading = loading

    D = shape.Triangle(10, 20, 30)
    @param(name='Enzyme loading', element=M301, kind='coupled', units='mg/g',
           baseline=20, distribution=D)
    def set_R301_enzyme_loading(loading):
        M301.enzyme_loading = loading

    # Enzymatic hydrolysis
    D = shape.Triangle(0, 24, 56)
    @param(name='Enzymatic hydrolysis time', element=R301, kind='coupled', units='hr',
           baseline=24, distribution=D)
    def set_R301_saccharification_time(tau):
        R301.tau_saccharification = tau

    D = shape.Triangle(0.75, 0.9, 0.948-1e-6)
    @param(name='Enzymatic hydrolysis glucan-to-glucose', element=R301, kind='coupled', units='%',
           baseline=0.9, distribution=D)
    def set_R301_glucan_conversion(X):
        R301.saccharification_rxns[2].X = X

    # Fermentation
    D = shape.Triangle(5, 10, 15)
    @param(name='CSL loading', element=R301, kind='coupled', units='g/L',
           baseline=10, distribution=D)
    def set_CSL_loading(loading):
        R301.CSL_loading = loading

    R302 = u.R302
    # 1e-6 is to avoid generating tiny negative flow (e.g., 1e-14)
    D = shape.Triangle(0.9, 0.95, 1-1e-6)
    @param(name='Seed train fermentation ratio', element=R302, kind='coupled', units='%',
           baseline=0.95, distribution=D)
    def set_ferm_ratio(ratio):
        R302.ferm_ratio = ratio

    D = shape.Triangle(0.55, 0.76, 0.93)
    @param(name='Lactic acid yield', element=R301, kind='coupled', units='g/g',
           baseline=0.76, distribution=D)
    def set_lactic_yield(lactic_yield):
        R301.target_yield = lactic_yield
        set_yield(lactic_yield, R301, R302)

    D = shape.Triangle(0.33, 0.89, 1.66)
    @param(name='Productivity', element=R301, kind='coupled', units='g/L/hr',
           baseline=0.89, distribution=D)
    def set_lactic_productivity(productivity):
        R301.productivity = productivity
        R302.productivity = productivity * R302.ferm_ratio

    D = shape.Triangle(0.004, 0.07, 0.32)
    @param(name='Acetic acid yield', element=R301, kind='coupled', units='g/g',
           baseline=0.07, distribution=D)
    def set_acetic_yield(acetic_yield):
        R301_X = R301.cofermentation_rxns.X
        R302_X = R302.cofermentation_rxns.X
        R301_X[1] = R301_X[4] = min(acetic_yield, 1-1e-6-R301_X[0]-R301_X[2])
        R302_X[1] = R302_X[4] = min(R301_X[1]*R302.ferm_ratio, 1-1e-6-R302_X[0]-R302_X[2])

    D = shape.Triangle(0.05, 0.07, 0.1)
    @param(name='Inoculum ratio', element=R301, kind='coupled', units='%',
           baseline=0.07, distribution=D)
    def set_inoculum_ratio(ratio):
        R301.inoculum_ratio = ratio

    ######################## Separation parameters ########################
    S402 = u.S402
    D = shape.Triangle(0.95, 0.995, 1)
    @param(name='Gypsum split', element=S402, kind='coupled', units='',
           baseline=0.995, distribution=D)
    def set_S402_gypsum_split(split):
        gypsum_index = S402.chemicals.index('Gypsum')
        S402.split[gypsum_index] = split

    R401 = u.R401
    D = baseline_triangle(1, 0.1)
    @param(name='Acidulation time', element=R401, kind='coupled', units='hr',
           baseline=1, distribution=D)
    def set_R401_tau(tau):
        R401.tau = tau

    R402 = u.R402
    D = baseline_triangle(1, 0.1)
    @param(name='Esterification conversion factor', element=R402, kind='coupled', units='',
           baseline=1, distribution=D)
    def set_R402_conversion_factor(factor):
        R402.X_factor = factor

    R403 = u.R403
    D = baseline_triangle(0.8, 0.1)
    @param(name='Hydrolysis conversion', element=R403, kind='coupled', units='%',
           baseline=0.8, distribution=D)
    def set_R403_conversion_factor(X):
        R403.hydrolysis_rxns.X[:] = X

    ######################## Separation parameters ########################
    D = baseline_uniform(0.8, 0.1)
    @param(name='boiler efficiency', element=CHP, kind='coupled', units='%',
           baseline=0.8, distribution=D)
    def set_boiler_efficiency(efficiency):
        CHP.B_eff = efficiency

    return model_dct


# -*- coding: utf-8 -*-
"""
Created on Tue Nov 30 22:30:49 2021

@author: yrc2
"""

from biorefineries import oilcane as oc

cfdiff = 3.3111 * (
    oc.GWP_characterization_factors['biodiesel']
    - oc.GWP_characterization_factors['biodiesel displacement'] 
)
configurations = ['O1', 'O2', 'O1*', 'O2*']
for name in configurations:
    df = oc.get_monte_carlo(name)
    file_name = oc.monte_carlo_file(name)
    df[oc.GWP_ethanol_displacement.index] += (
        df[oc.biodiesel_production.index] * cfdiff
        / df[oc.ethanol_production.index]
    )
    df.to_excel(file_name)
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

import biosteam as bst
bst.speed_up()

from . import (
    _chemicals,
    _utils,
    _settings,
    _units,
    _facilities,
    _tea,
    _processes,
    systems,
    )

from ._chemicals import chems
from ._utils import auom, CEPCI
from ._settings import price, CFs
from ._processes import update_settings
from .systems import *
from .models import *

def load_system(kind='SSCF'):
    if not kind in ('SSCF', 'SHF'):
        raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')
    global flowsheet, groups, teas, funcs, \
        lactic_sys, lactic_tea, feedstock, lactic_acid, \
        simulate_and_print, simulate_fermentation_improvement, \
        simulate_separation_improvement, simulate_operating_improvement

    flowsheet = getattr(systems, f'{kind}_flowsheet')
    groups = getattr(systems, f'{kind}_groups')
    teas = getattr(systems, f'{kind}_teas')
    funcs = getattr(systems, f'{kind}_funcs')

    update_settings(chems)
    bst.main_flowsheet.set_flowsheet(flowsheet)

    lactic_sys = flowsheet.system.lactic_sys
    lactic_tea = teas['lactic_tea']
    feedstock = flowsheet.stream.feedstock
    lactic_acid = flowsheet.stream.lactic_acid

    simulate_and_print = lambda : systems.simulate_and_print(kind)
    simulate_fermentation_improvement = \
        lambda: systems.simulate_fermentation_improvement(kind)
    simulate_separation_improvement = \
        lambda: systems.simulate_separation_improvement(kind)
    simulate_operating_improvement = \
        lambda: systems.simulate_operating_improvement(kind)
    return

load_system('SSCF')

# __all__ = (
#     flowsheet, groups, teas, funcs, \
#     lactic_sys, lactic_tea, feedstock, lactic_acid, \
#     simulate_and_print, simulate_fermentation_improvement, \
#     simulate_separation_improvement, simulate_operating_improvement,
#     *lactic_sys.units, # not seem to be working
#     )


# -*- coding: utf-8 -*-
"""
Created on Thu Nov  4 14:28:33 2021

@author: yrc2
"""

__all__ = (
    'GWP_characterization_factors',
    'set_GWPCF',
    'GWP',
)

GWP = 'GWP'

# All values in cradle-to-gate except for CH4, which is in cradle-to-grave
GWP_characterization_factors = { # Material GWP cradle-to-gate [kg*CO2*eq / kg]
    'sugarcane': 0.02931 * 0.30 / 0.25, # GREET, modified from moisture content of 0.75 to 0.70
    'sweet sorghum': 0.02821 * 0.30 / 0.25, # GREET, modified from moisture content of 0.75 to 0.70
    #'feedstock': 0.0607, # dry basis, Ecoinvent 2021
    # 'protease': 8.07, # Assume same as cellulase
    'cellulase': 8.07, # GREET
    'H3PO4': 1.00, # GREET
    'lime': 1.164, # GREET
    'pure-glycerol': 1.6678, # Ecoinvent, TRACI, market for glycerine, RoW; 
    'crude-glycerol': 0.36, # GREET
    'biodiesel': 1.13, # Soybean biodiesel GREET
    'DAP': 1.66, # GREET
    'CSL': 1.56, # GREET
    'HCl': 1.96, # GREET
    'NaOH': 2.01, # GREET
    'gasoline': 0.84, # GREET
    'methanol': 0.45, # GREET, Natural gas to methanol
    'NaOCH3': 1.5871, # Ecoinvent, TRACI, sodium methoxide
    'CH4': 0.33, # Natural gas from shell conventional recovery, GREET; includes non-biogenic emissions
    'Electricity': 0.36, # [kg*CO2*eq / kWhr] From GREET; NG-Fired Simple-Cycle Gas Turbine CHP Plant
    # 0.66 is the GWP from producing diesel from GREET; Conventional diesel from crude oil for US Refineries.
    # Downstream fuel emissions are added in. Accounts for how biodiesel has less energy than diesel.
    'biodiesel displacement': 0.92 * (0.66 +  (12 * 12.01 + 24 * 16) / (12 * 12.01 + 23 * 1.008)) 
}

GWP_characterization_factors['methanol catalyst mixture'] = (
    GWP_characterization_factors['methanol'] * 0.75 + GWP_characterization_factors['NaOH'] * 0.25
)

def set_GWPCF(stream, name, dilution=1.):
    stream.characterization_factors[GWP] = GWP_characterization_factors[name] * dilution
    
    
# from thermosteam.units_of_measure import convert
# from thermosteam import Chemical
# CH4 = Chemical('CH4')
# CO2 = Chemical('CO2')
# electricty_produced_per_kg_CH4 = - convert(0.8 * 0.85 * CH4.LHV / CH4.MW, 'kJ', 'kWhr')
# GWP_per_kg_CH4 = 0.33 + CO2.MW / CH4.MW
# GWP_per_kWhr = GWP_per_kg_CH4 / electricty_produced_per_kg_CH4


# -*- coding: utf-8 -*-
"""
Created on Mon Apr 19 03:19:53 2021

@author: yrc2
"""
from biosteam import Unit, BatchBioreactor
from thermosteam import PRxn, Rxn
from biorefineries.cornstover.units import (
    SeedTrain,
    CoFermentation
)

__all__ = ('SeedTrain', 'CoFermentation')

class SeedTrain(SeedTrain):
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None, reactions=None, saccharification=False):
        Unit.__init__(self, ID, ins, outs, thermo)
        self.saccharification = saccharification
        chemicals = self.chemicals
        self.reactions = reactions or PRxn([
    #   Reaction definition                   Reactant    Conversion
    Rxn('Glucose -> 2 Ethanol + 2 CO2',       'Glucose',   0.9000, chemicals),
    Rxn('3 Xylose -> 5 Ethanol + 5 CO2',      'Xylose',    0.8000, chemicals),
    Rxn('Glucose -> Cellmass',                'Glucose',  0.0473, chemicals),
    Rxn('Xylose -> Cellmass',                 'Xylose',  0.0421, chemicals),
        ])
        
    def _setup(self):
        self.outs[0].phase = 'g'
        
    def _run(self):
        vent, effluent= self.outs
        effluent.mix_from(self.ins, energy_balance=False)
        self.reactions.force_reaction(effluent)
        effluent.mol[effluent.mol < 0.] = 0.
        effluent.T = self.T
        vent.copy_flow(effluent, 'CO2', remove=True)
        

class CoFermentation(CoFermentation):
    
    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 tau=36, N=None, V=3785.4118, T=305.15, P=101325,
                 Nmin=2, Nmax=36, cofermentation=None):
        BatchBioreactor.__init__(self, ID, ins, outs, thermo, tau, N, V, T, P, Nmin, Nmax)
        self.P = P
        chemicals = self.chemicals
        self.loss = None
        self.cofermentation = cofermentation or PRxn([
    #   Reaction definition                   Reactant    Conversion
    Rxn('Glucose -> 2 Ethanol + 2 CO2',       'Glucose',   0.9500, chemicals),
    Rxn('3 Xylose -> 5 Ethanol + 5 CO2',      'Xylose',    0.8500, chemicals),
    Rxn('Glucose -> Cellmass',                'Glucose',  0.05, chemicals),
    Rxn('Xylose -> Cellmass',                 'Xylose',  0.05, chemicals),
        ])
    
        self.CSL_to_constituents = Rxn(
            'CSL -> 0.5 H2O + 0.25 LacticAcid + 0.25 Protein', 'CSL', 1.0000, chemicals, basis='wt',
        )
        self.CSL_to_constituents.basis = 'mol'
        
        if all([i in self.chemicals for i in ('FFA', 'DAG', 'TAG', 'Glycerol')]):
            self.lipid_reaction = self.oil_reaction = PRxn([
                Rxn('TAG + 3Water -> 3FFA + Glycerol', 'TAG', 0.23, chemicals),
                Rxn('TAG + Water -> FFA + DAG', 'TAG', 0.02, chemicals)
            ])
        else:
            self.lipid_reaction = self.oil_reaction = None
            


# -*- coding: utf-8 -*-
"""
Created on Tue Dec  7 06:20:39 2021

@author: yrc2
"""
import biosteam as bst
from biorefineries import oilcane as oc

bst.Metric.caching = False
oc.load('O1')
model = oc.model
names = (
    'Bagasse oil retention',
    'Oil extraction efficiency',
    'Plant capacity',
    'Ethanol price',
    'Relative biodiesel price',
    'Natural gas price',
    'Electricity price',
    'Operating days',
    'IRR',
    'Crude glycerol price',
    'Pure glycerol price',
    'Saccharification reaction time',
    'Cellulase price',
    'Cellulase loading',
    'PTRS base cost',
    'Cane glucose yield',
    'Sorghum glucose yield',
    'Cane xylose yield',
    'Sorghum xylose yield',
    'Glucose to ethanol yield',
    'Xylose to ethanol yield',
    'Titer',
    'Productivity',
    'Cane PL content',
    'Sorghum PL content',
    'Cane FFA content',
    'Sorghum FFA content',
    'Cane oil content',
    'Relative sorghum oil content',
    'TAG to FFA conversion',
    'Feedstock GWPCF',
    'Methanol GWPCF',
    'Pure glycerine GWPCF',
    'Cellulase GWPCF',
    'Natural gas GWPCF',
)
for name, parameter in zip(names, model.parameters): parameter.name = name

oil_related = {
    'Bagasse oil retention',
    'Oil extraction efficiency',
    'Plant capacity',
    'Ethanol price',
    'Relative biodiesel price',
    'Electricity price',
    'Operating days',
    'IRR',
    'Crude glycerol price',
    'Pure glycerol price',
    'Cane PL content',
    'Cane FFA content',
    'Cane oil content',
    'TAG to FFA conversion',
    'Feedstock GWPCF',
    'Methanol GWPCF',
    'Pure glycerine GWPCF',
}
model.parameters = [p for p in model.parameters if p.name in oil_related]
        
names = (
    (0, 'Maximum feedstock purchase price'),
    (2, None),
    (3, None),
    (4, None),
    (5, None),
    (9, 'Ethanol GWP, economic allocation'),
    (10, 'Biodiesel GWP, economic allocation'),
    (11, 'Crude glycerol GWP, economic allocation'),
    (12, 'Electricity GWP, economic allocation'),
    (13, 'Ethanol GWP, displacement allocation'),
    (15, 'Ethanol GWP, energy allocation'),
    (16, 'Biodiesel GWP, energy allocation'),
    (17, 'Crude glycerol GWP, energy allocation'),
)

def rename(metric, name):
    if name is not None: metric.name = name
    return metric

model.metrics = [rename(model.metrics[index], name) for index, name in names]


# -*- coding: utf-8 -*-
"""
Created on Fri May 14 22:48:34 2021

@author: yrc2
"""
from matplotlib import pyplot as plt
import numpy as np
from chaospy import distributions as shape
from math import sqrt

__all__ = (
    'ethanol_price_distribution', 
    'biodiesel_price_distribution',
    'biodiesel_minus_ethanol_price_distribution',
    'natural_gas_price_distribution',
    'mean_glycerol_price',
    'mean_ethanol_price',
    'mean_biodiesel_price',
    'mean_natural_gas_price',
    'mean_electricity_price',
    'mean_soymeal_price',
)

ethanol_prices = [ # USD / gal Dec 2009 - Nov 2020, by quarter
    2.08, 1.66, 1.71, 2.33, 2.37, 2.67, 2.87, 2.83, 2.30, 2.26, 2.47,
    2.48, 2.39, 2.58, 2.57, 2.31, 2.19, 2.76, 2.28, 2.16, 1.83, 1.63, 1.64,
    1.61, 1.47, 1.52, 1.59, 1.58, 1.57, 1.60, 1.62, 1.24, 1.02, 1.30, 1.37, 
    1.20, 1.09, 1.09, 1.32, 1.35
]

biodiesel_prices = [ # USD / gal Dec 2009 - Nov 2020, by month
    3.38, 3.27, 3.3, 3.32, 3.33, 3.21, 3.18, 3.18, 3.3, 
    3.35, 3.66, 3.9, 4.25, 4.49, 4.6, 4.82, 5.07, 5.23, 5.4, 5.45, 5.44, 5.74, 
    5.58, 5.34, 4.92, 4.67, 4.67, 4.79, 4.97, 4.76, 4.4, 4.39, 4.47, 4.53,
    4.49, 4.12, 3.94, 4.37, 4.51, 4.73, 4.75, 4.84, 4.85, 5.01, 4.9, 4.84,
    4.56, 3.96, 3.9, 3.39, 3.37, 3.67, 3.75, 3.76, 3.7, 3.66, 3.48, 3.36,
    3.14, 3.15, 3.03, 2.83, 2.88, 2.9, 2.89, 3.08, 3.18, 2.96, 2.55, 2.6,
    2.67, 2.65, 2.84, 2.79, 2.92, 3.17, 3.21, 3.28, 3.31, 3.11, 3.23, 3.35,
    3.37, 3.49, 3.64, 3.52, 3.45, 3.14, 3.11, 3.18, 3.13, 3.23, 3.31, 3.33,
    3.3, 3.34, 3.23, 3.45, 3.16, 3.04, 3.02, 3.13, 3.11, 3.02, 3.01, 3.04,
    3.09, 3.09, 3.06, 3.12, 3.24, 3.08, 2.98, 2.86, 2.87, 2.91, 2.99, 3.06, 
    3.13, 3.21, 
]


N_quarters = int(len(biodiesel_prices)/3)
biodiesel_prices_quarter = [ # By quarter
    np.mean(biodiesel_prices[3*i: 3*i+3]) for i in range(N_quarters)
]

biodiesel_minus_ethanol_prices = [i - j for i, j in zip(biodiesel_prices, ethanol_prices)]

natural_gas_prices = [ # City gate [USD / cf] 2010 to 2019
    6.18, 5.63, 4.73, 4.88, 5.71, 4.26, 3.71, 4.16, 4.23, 3.81,
]

electricity_prices = []
def triangular_distribution(x):
    a, b, c = fit_triangular_distribution(x)
    return shape.Triangle(a, c, b)

def triangular_distribution_median(a, b, c):
    if c >= (a + b) / 2: 
        return a + sqrt((b - a) * (c - a) * 0.5)
    else:
        return b - sqrt((b - a) * (b - c) * 0.5)

def fit_triangular_distribution(x, median=False):
    a = min(x)
    b = max(x)
    if median:
        median = np.median(x)
        if median >= (a + b) / 2: 
            dummy = a - median
            c = a + 2 * dummy * dummy / (b - a)
        else:
            dummy = b - median
            c = b - 2 * dummy * dummy / (b - a)
        assert triangular_distribution_median(a, b, c) == median 
    else:
        c = 3 * np.mean(x) - (a + b)
        if c < a: c = a + 1e-6
    return a, b, c

def plot_triangular_distribution(a, b, c):
    return plt.plot([a, c, b], [0, 2 / (b - a), 0])

def plot_histogram(x, *args, bins=10, density=True, **kwargs):
    return plt.hist(x, *args, **kwargs)


# https://www.eia.gov/outlooks/aeo/pdf/00%20AEO2021%20Chart%20Library.pdf
# Data from historical prices, 2010-2020
electricity_price_distribution = shape.Triangle(0.0583, 0.065, 0.069)
ethanol_price_distribution = triangular_distribution(ethanol_prices)
biodiesel_price_distribution = triangular_distribution(biodiesel_prices)
biodiesel_minus_ethanol_price_distribution = triangular_distribution(biodiesel_minus_ethanol_prices)
natural_gas_price_distribution = triangular_distribution(natural_gas_prices)

mean_glycerol_price = (0.10 + 0.22) * 0.5 
mean_ethanol_price = np.mean(ethanol_prices)
mean_biodiesel_price = np.mean(biodiesel_prices)
mean_natural_gas_price = np.mean(natural_gas_prices)
mean_electricity_price = sum([0.0583, 0.065, 0.069]) / 3.
mean_soymeal_price = 0.33 # 10 yr average; https://markets.businessinsider.com/commodities/soybean-meal-price?op=1


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
#
# This module is under the UIUC open-source license. See
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %%

import biosteam as bst
from ._chemicals import chems
from ._processes import (
    update_settings,
    create_preprocessing_process,
    create_pretreatment_process,
    create_SSCF_conversion_process,
    create_SHF_conversion_process,
    create_separation_process,
    create_wastewater_process,
    create_facilities,
    create_lactic_sys
    )

bst.speed_up()
update_settings(chems)

__all__ = (
    'create_SSCF_sys', 'create_SHF_sys',
    'SSCF_flowsheet', 'SSCF_groups', 'SSCF_teas', 'SSCF_funcs',
    'SHF_flowsheet', 'SHF_groups', 'SHF_teas', 'SHF_funcs',
    'simulate_and_print', 'simulate_separation_improvement',
    'simulate_separation_improvement', 'simulate_operating_improvement'
    )


# %%

def create_SSCF_sys():
    flowsheet = bst.Flowsheet('SSCF')
    s = flowsheet.stream
    u = flowsheet.unit

    flowsheet, groups, get_feedstock_dry_mass, get_flow_tpd = \
        create_preprocessing_process(flowsheet)

    flowsheet, groups = \
        create_pretreatment_process(flowsheet, groups, u.U101-0, get_feedstock_dry_mass)

    flowsheet, groups = \
        create_SSCF_conversion_process(flowsheet, groups, u.P201-0)

    flowsheet, groups = \
        create_separation_process(flowsheet, groups, u.PS301-0, kind='SSCF')

    # The last one is reserved for blowdown
    WWT_streams = (u.H201-0, u.M401_P-0, u.R402-1, u.R403-1, '')
    flowsheet, groups = \
        create_wastewater_process(flowsheet, groups, get_flow_tpd, WWT_streams)

    CHP_wastes = (u.U101-1, u.S401-0, u.S504-1)
    CHP_biogas = u.R501-0
    CHP_side_streams = (s.water_M201, s.water_M202, s.steam_M203)
    process_water_streams = {
        'pretreatment': (s.water_M201, s.water_M202, s.steam_M203, s.water_M205),
        'conversion': (s.water_M301, s.water_R301),
        'separation': (s.water_R403,)
        }
    recycled_water = u.S505-0
    flowsheet, groups = \
        create_facilities(flowsheet, groups, get_flow_tpd,
                          CHP_wastes, CHP_biogas, CHP_side_streams,
                          process_water_streams, recycled_water)

    flowsheet, teas, funcs = create_lactic_sys(flowsheet, groups, get_flow_tpd)

    return flowsheet, groups, teas, funcs

def create_SHF_sys():
    flowsheet = bst.Flowsheet('SHF')
    s = flowsheet.stream
    u = flowsheet.unit

    flowsheet, groups, get_feedstock_dry_mass, get_flow_tpd = \
        create_preprocessing_process(flowsheet)

    flowsheet, groups = \
        create_pretreatment_process(flowsheet, groups, u.U101-0, get_feedstock_dry_mass)

    flowsheet, groups = \
        create_SHF_conversion_process(flowsheet, groups, u.P201-0)

    flowsheet, groups = \
        create_separation_process(flowsheet, groups, u.PS301-0, kind='SHF')

    # The last one is reserved for blowdown
    WWT_streams = (u.H201-0, u.E301-1, u.M401_P-0, u.R402-1, u.R403-1, '')
    flowsheet, groups = \
        create_wastewater_process(flowsheet, groups, get_flow_tpd, WWT_streams)

    CHP_wastes = (u.U101-1, u.S301-0, u.S401-0, u.S504-1)
    CHP_biogas = u.R501-0
    CHP_side_streams = (s.water_M201, s.water_M202, s.steam_M203)
    process_water_streams = {
        'pretreatment': (s.water_M201, s.water_M202, s.steam_M203, s.water_M205),
        'conversion': (s.water_M301, s.water_R301),
        'separation': (s.water_R403,)
        }
    recycled_water = u.S505-0
    flowsheet, groups = \
        create_facilities(flowsheet, groups, get_flow_tpd,
                          CHP_wastes, CHP_biogas, CHP_side_streams,
                          process_water_streams, recycled_water)

    flowsheet, teas, funcs = create_lactic_sys(flowsheet, groups, get_flow_tpd)

    return flowsheet, groups, teas, funcs

SSCF_flowsheet, SSCF_groups, SSCF_teas, SSCF_funcs = create_SSCF_sys()
SHF_flowsheet, SHF_groups, SHF_teas, SHF_funcs  = create_SHF_sys()


# %%

# =============================================================================
# Useful functions for summarizing results and considering alternative process
# decision variables
# =============================================================================

def simulate_and_print(system='SSCF'):
    if 'SSCF' in str(system).upper():
        flowsheet = SSCF_flowsheet
        funcs = SSCF_funcs
    elif 'SHF' in str(system).upper():
        flowsheet = SHF_flowsheet
        funcs = SHF_funcs
    else:
        raise ValueError(f'system can only be "SSCF" or "SHF", not {system}.')
    bst.main_flowsheet.set_flowsheet(flowsheet)

    print('\n---------- Simulation Results ----------')
    print(f'MPSP is ${funcs["simulate_get_MPSP"]():.3f}/kg')
    print(f'GWP is {funcs["get_GWP"]():.3f} kg CO2-eq/kg lactic acid')
    print(f'FEC is {funcs["get_FEC"]():.2f} MJ/kg lactic acid')
    print('------------------------------------------\n')


def simulate_fermentation_improvement(kind='SSCF'):
    if 'SSCF' in str(kind).upper():
        flowsheet = SSCF_flowsheet
    elif 'SHF' in str(kind).upper():
        flowsheet = SHF_flowsheet
    else:
        raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')
    bst.main_flowsheet.set_flowsheet(flowsheet)
    u = flowsheet.unit
    flowsheet.system.lactic_sys.simulate()

    R301_X = u.R301.cofermentation_rxns.X
    R302_X = u.R302.cofermentation_rxns.X
    u.R301.target_yield = 0.95
    R301_X[0] = R301_X[3] = 0.95
    R301_X[1] = R301_X[4] = 0
    R302_X[1] = R302_X[4] = 0
    simulate_and_print(kind)

def simulate_separation_improvement(kind='SSCF'):
    if 'SSCF' in str(kind).upper():
        flowsheet = SSCF_flowsheet
    elif 'SHF' in str(kind).upper():
        flowsheet = SHF_flowsheet
    else:
        raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')
    bst.main_flowsheet.set_flowsheet(flowsheet)
    u = SHF_flowsheet.unit
    flowsheet.system.lactic_sys.simulate()

    u.R402.X_factor = 0.9/u.R402.esterification_rxns.X[0]
    u.R403.hydrolysis_rxns.X[:] = 0.9
    simulate_and_print(kind)

def simulate_operating_improvement(kind='SSCF'):
    if 'SSCF' in str(kind).upper():
        flowsheet = SSCF_flowsheet
        funcs = SSCF_funcs
    elif 'SHF' in str(kind).upper():
        flowsheet = SHF_flowsheet
        funcs = SHF_funcs
    else:
        raise ValueError(f'kind can only be "SSCF" or "SHF", not {kind}.')
    bst.main_flowsheet.set_flowsheet(flowsheet)
    s = flowsheet.stream
    u = flowsheet.unit
    flowsheet.system.lactic_sys.simulate()

    u.U101.diversion_to_CHP = 0.25
    print('\n---------- Simulation Results ----------')
    print(f'MPSP is ${funcs["simulate_get_MPSP"]():.3f}/kg')
    s.LCA_stream.imass['CH4'] *= 0.75
    s.natural_gas.imass['CH4'] *= 0.75
    print(f'GWP is {funcs["get_GWP"]():.3f} kg CO2-eq/kg lactic acid')
    print(f'FEC is {funcs["get_FEC"]():.2f} MJ/kg lactic acid')
    print('------------------------------------------\n')


# -*- coding: utf-8 -*-
"""
Web-app model for AWS lambda

"""

import biosteam as bst
from chaospy import distributions as shape
import biorefineries.cornstover as cs

sys = cs.cornstover_sys
tea = cs.cornstover_tea
model = bst.Model(sys, exception_hook='raise')

# =============================================================================
# Define metrics
# =============================================================================

metric = model.metric
kg_per_ton = 907.185
ethanol_density_kggal = cs.ethanol_density_kggal
cornstover = cs.cornstover
ethanol = cs.ethanol
pretreatment_conversions = cs.R201.reactions.X
cofermentation_conversions = cs.R303.cofermentation.X
saccharification_conversions = cs.R303.saccharification.X
BT = cs.BT

@metric(name='Minimum ethanol selling price', units='USD/gal')
def get_MESP(): 
    return tea.solve_price(ethanol) * ethanol_density_kggal

# Convert from kg/hr to MM gal/yr
@metric(name='Ethanol production', units='10^6*gal/yr')
def get_ethanol_production():
    return ethanol.F_mass / ethanol_density_kggal * tea.operating_hours / 1e6

# Yield in gal/dry-U.S.ton
@metric(name='Ethanol yield', units='gal/dry US ton')
def get_ethanol_yield():
    ethanol_galyr = ethanol.F_mass / ethanol_density_kggal
    dry_feedstock_tonyr = (cornstover.F_mass - cornstover.imass['H2O']) / kg_per_ton
    return ethanol_galyr / dry_feedstock_tonyr

# Total capital investment in MM$
@metric(name='Total capital investment', units='10^6*USD')
def get_TCI():
    return tea.TCI / 1e6

# Annual operating cost
@metric(name='Annual operating cost', units='10^6*USD/yr')
def get_AOC():
    return tea.AOC / 1e6

@metric(name='Net electricity production', units='MWhr/yr')
def get_net_electricity():
    return -sys.power_utility.rate * sys.operating_hours / 1e3

@metric(name='Electricity credit', units='10^6*USD/yr')
def get_electricity_credit():
    return -tea.utility_cost / 1e6

# =============================================================================
# Define parameters
# =============================================================================

def param(name, baseline, bounds=None, **kwargs):
    lb = 0.9 * baseline
    ub = 1.1 * baseline
    if bounds is not None:
        if lb < bounds[0]:
            lb = bounds[0]
        if ub > bounds[1]:
            ub = bounds[1]
    distribution = shape.Uniform(lb, ub)
    return model.parameter(name=name, bounds=bounds, distribution=distribution, baseline=baseline, **kwargs)

@param(name='Cornstover price', element=cornstover, kind='isolated', 
       units='USD/ton', baseline=cornstover.price * kg_per_ton)
def set_cornstover_price(price):
    cornstover.price = price / kg_per_ton

@param(name='Enzyme price', element=cs.cellulase, kind='isolated',
       description='price of cellulase enzyme mixture containing 50 g of cellulase per 1000g of cocktail',
       units='$USD/ton', baseline=cs.cellulase.price * kg_per_ton)
def set_cellulase_price(price):
    cs.cellulase.price = price / kg_per_ton

@param(name='Electricity price', element='TEA', kind='isolated', units='USD/kWh',
       baseline=bst.PowerUtility.price)
def set_electricity_price(price):
    bst.PowerUtility.price = price    

@param(name='Income tax rate', element='TEA', kind='isolated', units='%',
       baseline=tea.income_tax * 100)
def set_tax_rate(rate):
    tea.income_tax = rate / 100

@param(name='Plant capacity', element=cornstover, kind='coupled', units='dry US ton/yr',
       baseline=(cornstover.F_mass - cornstover.imass['H2O']) * tea.operating_hours / kg_per_ton,
       description="annual feestock processing capacity")
def set_plant_size(flow_rate):
    dry_content = 1 - cornstover.imass['H2O'] / cornstover.F_mass
    cornstover.F_mass = flow_rate / tea.operating_hours / dry_content * kg_per_ton

@param(name='PT glucan-to-glucose', element=cs.R201, kind='coupled', units='% theoretical',
       description='extent of reaction, glucan + water -> glucose, in pretreatment reactor',
       baseline=pretreatment_conversions[0] * 100,
       bounds=(0, 100))
def set_PT_glucan_to_glucose(X):
    X /= 100.
    pretreatment_conversions[0] = X
    corxns = pretreatment_conversions[1:3] 
    corxns[:] = 0.003
    if pretreatment_conversions[:3].sum() > 1.:
        f = corxns / corxns.sum()
        corxns[:] = f * (1. - X) * 0.9999999

@param(name='PT xylan-to-xylose', element=cs.R201, kind='coupled', units='% theoretical',
       description='extent of reaction, xylan + water -> xylose, in pretreatment reactor',
       baseline=pretreatment_conversions[8] * 100,
       bounds=(0, 100))
def set_PT_xylan_to_xylose(X):
    X /= 100.
    pretreatment_conversions[8] = X
    corxns = pretreatment_conversions[9:11] 
    corxns[:] = [0.024, 0.05]
    if pretreatment_conversions[8:11].sum() > 1.:
        f = corxns / corxns.sum()
        corxns[:] = f * (1. - X) * 0.9999999

@param(name='PT xylan-to-furfural', element=cs.R201, kind='coupled', units='% theoretical',
       description='extent of reaction, xylan -> furfural + 2 water, in pretreatment reactor',
       baseline=pretreatment_conversions[10] * 100,
       bounds=(0, 100))
def set_PT_xylan_to_furfural(X):
    # To make sure the overall xylan conversion doesn't exceed 100%
    lb = 1. - pretreatment_conversions[8] - pretreatment_conversions[9]
    pretreatment_conversions[10] = min(lb, X / 100.) * 0.9999999

@param(name='EH cellulose-to-glucose', element=cs.R303, kind='coupled', units='% theoretical',
       description='extent of reaction, gluan + water -> glulose, in enzyme hydrolysis',
       baseline=saccharification_conversions[2] * 100,
       bounds=(0, 100))
def set_EH_glucan_to_glucose(X):
    X /= 100.
    saccharification_conversions[2] = X
    corxns = saccharification_conversions[:2] 
    corxns[:] = [0.04, 0.0012]
    if saccharification_conversions[:3].sum() > 1.:
        f = corxns / corxns.sum()
        corxns[:] = f * (1. - X) * 0.9999999

@param(name='FERM glucose-to-ethanol', element=cs.R303, kind='coupled', units='% theoretical',
       description='extent of reaction, glucose -> 2 ethanol + 2 CO2, in enzyme hydrolysis',
       baseline=cofermentation_conversions[0] * 100,
       bounds=(0, 100))
def set_FERM_glucose_to_ethanol(X):
    X /= 100.
    cofermentation_conversions[0] = X
    corxns = cofermentation_conversions[1:4] 
    corxns[:] = [0.02, 0.0004, 0.006]
    if cofermentation_conversions[:4].sum() > 1.:
        f = corxns / corxns.sum()
        corxns[:] = f * (1. - X) * 0.9999999

@param(name='Boiler efficiency', element=BT, kind='coupled', units='%',
       description='efficiency of burning fuel to produce steam',
       baseline=BT.boiler_efficiency * 100,
       bounds=(0, 100))
def set_boiler_efficiency(X):
    BT.boiler_efficiency = X / 100.

@param(name='Turbogenerator efficiency', element=BT, kind='coupled', units='%',
       description='efficiency of converting steam to power',
       baseline=BT.turbogenerator_efficiency * 100,
       bounds=(0, 100))
def set_turbogenerator_efficiency(X):
    BT.turbogenerator_efficiency = X / 100.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2021, Yoel Cortes-Pena <yoelcortes@gmail.com>
# Bioindustrial-Park: BioSTEAM's Premier Biorefinery Models and Results
# Copyright (C) 2020-2021, Yalin Li <yalinli2@illinois.edu>,
# Sarang Bhagwat <sarangb2@illinois.edu>, and Yoel Cortes-Pena (this biorefinery)
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.


# %% 

# =============================================================================
# Setup
# =============================================================================

import numpy as np
import pandas as pd
import biosteam as bst
from biosteam.utils import TicToc
from biorefineries.lactic import load_system, \
    SSCF_flowsheet, SSCF_funcs, SHF_flowsheet, SHF_funcs
from biorefineries.lactic._utils import _feedstock_factor

bst.speed_up()


# %%

# =============================================================================
# Model to evalute system across feedstock carbohydate content at different
# feedstock price metrics
# =============================================================================

carbs = ('Glucan', 'Xylan', 'Arabinan', 'Galactan', 'Mannan')

def set_carbs(carbs_content, feedstock):
    dry_mass = feedstock.F_mass.copy() - feedstock.imass[('H2O',)].copy()
    old_carbs_mass_total = feedstock.imass[carbs].sum().copy()
    ratio = feedstock.get_normalized_mass(carbs)
    new_carbs_mass = dry_mass * carbs_content * ratio
    feedstock.set_flow(new_carbs_mass, 'kg/hr', carbs)
    mass_diff = new_carbs_mass.sum() - old_carbs_mass_total
    feedstock.imass['Extractives'] -= mass_diff
    if any(feedstock.mass < 0):
        raise ValueError(f'Carbohydrate content of {carbs_content*100:.1f} dw% is infeasible')

get_feedstock_carbs = \
    lambda: feedstock.imass[carbs].sum()/(feedstock.F_mass-feedstock.imass['H2O'])

prices = np.arange(0, 210, 10)


# %%

# =============================================================================
# Evaluate across feedstock price and carbohydrate content
# =============================================================================

# Initiate a timer
timer = TicToc('timer')
timer.tic()

TEA_set_carbs = []
TEA_actual_carbs = []
TEA_prices = []
TEA_titers = []
TEA_yields = []
LCA_actual_carbs = []
LCA_titers = []
LCA_yields = []
MPSPs = []
NPVs = []
GWPs = []
FECs = []

# Configuration 2
load_system('SHF')
carb_contents1 = np.arange(0.25, 0.59, 0.01)
carb_contents1 = carb_contents1.tolist() + [0.589]
R301 = SHF_flowsheet.unit.R301
R301.allow_dilution = False
R301.allow_concentration = True
R301.mode = 'batch'
R301.feed_freq = 1
R301.target_titer = 97.5

feedstock = SHF_flowsheet.stream.feedstock
lactic_acid = SHF_flowsheet.stream.lactic_acid
lactic_sys = SHF_flowsheet.system.lactic_sys
lactic_tea = lactic_sys.TEA

# Using two loops are not optimal, can potentially use Model and Metric to speed up
for i in carb_contents1:
    set_carbs(i, feedstock)
    LCA_actual_carbs.append(get_feedstock_carbs())    
    lactic_sys.simulate()
    LCA_titers.append(R301.effluent_titer)
    LCA_yields.append(R301.lactic_yield)
    GWPs.append(SHF_funcs['get_GWP']())
    FECs.append(SHF_funcs['get_FEC']())
    for j in prices:
        TEA_set_carbs.append(i)
        TEA_actual_carbs.append(get_feedstock_carbs())
        TEA_prices.append(j)
        TEA_titers.append(R301.effluent_titer)
        TEA_yields.append(R301.lactic_yield)
        feedstock.price = j / _feedstock_factor
        lactic_acid.price = 0
        for m in range(3):
            MPSP = lactic_acid.price = lactic_tea.solve_price(lactic_acid)
        MPSPs.append(MPSP)
        NPVs.append(lactic_tea.NPV)

# Then concentration needed to get to the baseline titer
load_system('SSCF')
carb_contents2 = np.arange(0.59, 0.701, 0.01).tolist()
R301 = SSCF_flowsheet.unit.R301
R301.allow_dilution = True
R301.allow_concentration = False
R301.mode = 'batch'
R301.feed_freq = 1
R301.target_titer = 97.5

feedstock = SSCF_flowsheet.stream.feedstock
lactic_acid = SSCF_flowsheet.stream.lactic_acid
lactic_sys = SSCF_flowsheet.system.lactic_sys
lactic_tea = lactic_sys.TEA

for i in carb_contents2:
    set_carbs(i, feedstock)
    LCA_actual_carbs.append(get_feedstock_carbs())
    lactic_sys.simulate()
    LCA_titers.append(R301.effluent_titer)
    LCA_yields.append(R301.lactic_yield)
    GWPs.append(SSCF_funcs['get_GWP']())
    FECs.append(SSCF_funcs['get_FEC']())
    for j in prices:
        TEA_set_carbs.append(i)
        TEA_actual_carbs.append(get_feedstock_carbs())
        TEA_prices.append(j)
        TEA_titers.append(R301.effluent_titer)
        TEA_yields.append(R301.lactic_yield)        
        feedstock.price = j / _feedstock_factor
        lactic_acid.price = 0
        for m in range(3):
            MPSP = lactic_acid.price = lactic_tea.solve_price(lactic_acid)
        MPSPs.append(MPSP)
        NPVs.append(lactic_tea.NPV)

TEA_data = pd.DataFrame({
    'Target carb content [dw%]': TEA_set_carbs,
    'Actual carb content [dw%]': TEA_actual_carbs,
    'Price [$/dry-ton]': TEA_prices,
    'Titer [g/L]': TEA_titers,
    'Yield [dw%]': TEA_yields,
    'MPSP [$/kg]': MPSPs,
    'NPVs [$]': NPVs
    })

LCA_data = pd.DataFrame({
    'Target carb content [dw%]': carb_contents1+carb_contents2,
    'Actual carb content [dw%]': LCA_actual_carbs,
    'Titer [g/L]': LCA_titers,
    'Yield [dw%]': LCA_yields,
    'GWP [kg CO2-eq/kg]': GWPs,
    'FEC [MJ/kg]': FECs
    })

'''Output to Excel'''
with pd.ExcelWriter('3_feedstock_carb-price.xlsx') as writer:
    TEA_data.to_excel(writer, sheet_name='TEA')
    LCA_data.to_excel(writer, sheet_name='LCA')

print(f'\nSimulation time is {timer.elapsed_time/60:.1f} min')















# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from . import (utils,
               _process_settings,
               _chemicals,
               _system,
               _tea,
)

__all__ = [*utils.__all__,
           *_process_settings.__all__,
           *_chemicals.__all__,
           *_system.__all__,
           *_tea.__all__,
           'lipidcane_sys',
           'lipidcane_tea', 
           'flowsheet',
]

from .utils import *
from ._process_settings import *
from ._chemicals import *
from ._system import *
from ._tea import *

_system_loaded = False
_chemicals_loaded = False

def load():
    if not _chemicals_loaded: _load_chemicals()
    _load_system(globals())

def _load_chemicals():
    global chemicals, _chemicals_loaded
    chemicals = create_chemicals()
    _chemicals_loaded = True

def _load_system(dct):
    import biosteam as bst
    from biosteam import main_flowsheet as F, UnitGroup
    global lipidcane_sys, lipidcane_tea, specs, flowsheet, _system_loaded
    global biodiesel_production_units, ethanol_production_units
    flowsheet = bst.Flowsheet('lipidcane')
    F.set_flowsheet(flowsheet)
    bst.settings.set_thermo(chemicals)
    load_process_settings()
    lipidcane_sys = create_lipidcane_to_biodiesel_and_conventional_ethanol_system()
    lipidcane_sys.simulate()
    lipidcane_tea = create_tea(lipidcane_sys)
    lipidcane_tea.IRR = lipidcane_tea.solve_IRR()
    dct.update(flowsheet.to_dict())
    biodiesel_production_units = UnitGroup('Biodisel production units', 
                                           [T401, P401, T402, P402, T403,
                                            P403, T404, P404, S401, R401,
                                            C401, P405, R402, R402, C402,
                                            T405, P406, C403, F401, P407,
                                            H401, P408, T406, P409, C404,
                                            T407, P410, D401, H402, D402,
                                            H403, P411, H404, P412, T408,
                                            T409])
    ethanol_production_units = UnitGroup('Ethanol production units',
                                         [S301, F301, P306, M301, H301,
                                          T305, R301, T301, C301, D301,
                                          M302, P301, H302, D302, P302,
                                          M303, D303, H303, U301, H304,
                                          T302, P304, T303, P305, M304,
                                          T304, P303])
    _system_loaded = True

def __getattr__(name):
    if not _chemicals_loaded:
        _load_chemicals()
        if name == 'chemicals': return chemicals
    if not _system_loaded: 
        dct = globals()
        _load_system(dct)
        if name in dct: return dct[name]
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")



# -*- coding: utf-8 -*-
"""
Created on Fri Nov  5 00:04:53 2021

@author: yrc2
"""
import numpy as np
import biosteam as bst
from warnings import warn
from biorefineries import oilcane as oc
from ._variable_mockups import (
    all_metric_mockups
)
from ._parse_configuration import (
    parse,
    Configuration,
)
from ._load_data import (
    get_monte_carlo,
    monte_carlo_file,
    autoload_file_name,
    spearman_file,
)

__all__ = (
    'evaluate_configurations_across_extraction_efficiency_and_oil_content',
    'evaluate_configurations_across_sorghum_and_cane_oil_content',
    'evaluate_MFPP_uncertainty_across_ethanol_and_biodiesel_prices',
    'evaluate_MFPP_benefit_uncertainty_across_ethanol_and_biodiesel_prices',
    'evaluate_MFPP_benefit_across_ethanol_and_biodiesel_prices',
    'evaluate_MFPP_across_ethanol_and_biodiesel_prices',
    'evaluate_MFPP_benefit_across_ethanol_and_biodiesel_prices',
    'run_uncertainty_and_sensitivity',
    'run_all',
)

def no_derivative(f):
    def f_derivative_disabled(*args, **kwargs):
        oc.disable_derivative()
        try:
            return f(*args, **kwargs)
        finally:
            oc.enable_derivative()
    return f

def evaluate_configurations_across_extraction_efficiency_and_oil_content(
        efficiency, oil_content, oil_retention, agile, configurations,
    ):
    A = len(agile)
    C = len(configurations)
    M = len(all_metric_mockups)
    data = np.zeros([A, C, M])
    for ia in range(A):
        for ic in range(C):    
            oc.load([int(configurations[ic]), agile[ia]])
            if agile[ia]:
                oc.cane_mode.oil_content = oc.sorghum_mode.oil_content = oil_content
                oc.oil_extraction_specification.load_efficiency(efficiency)
                oc.oil_extraction_specification.load_oil_retention(oil_retention)
            else:
                oc.oil_extraction_specification.load_specifications(
                    efficiency=efficiency, 
                    oil_content=oil_content, 
                    oil_retention=oil_retention
                )
            oc.sys.simulate()
            data[ia, ic, :] = [j() for j in oc.model.metrics]
    return data

N_metrics = len(all_metric_mockups)
evaluate_configurations_across_extraction_efficiency_and_oil_content = no_derivative(
    np.vectorize(
        evaluate_configurations_across_extraction_efficiency_and_oil_content, 
        excluded=['oil_retention', 'agile', 'configurations'],
        signature=f'(),(),(),(a),(c)->(a,c,{N_metrics})'
    )
)
def evaluate_configurations_across_sorghum_and_cane_oil_content(
        sorghum_oil_content, cane_oil_content, configurations, relative,
    ):
    C = len(configurations)
    M = len(all_metric_mockups)
    data = np.zeros([C, M])
    for ic in range(C):
        oc.load([int(configurations[ic]), True])
        oc.cane_mode.oil_content = cane_oil_content
        if relative:
            oc.sorghum_mode.oil_content = cane_oil_content + sorghum_oil_content
        else:
            oc.sorghum_mode.oil_content = sorghum_oil_content
        oc.sys.simulate()
        data[ic, :] = [j() for j in oc.model.metrics]
    return data

evaluate_configurations_across_sorghum_and_cane_oil_content = no_derivative(
    np.vectorize(
        evaluate_configurations_across_sorghum_and_cane_oil_content, 
        excluded=['configurations', 'relative'],
        signature=f'(),(),(c),()->(c,{N_metrics})'
    )
)              

def evaluate_MFPP_uncertainty_across_ethanol_and_biodiesel_prices(name, ethanol_price, biodiesel_price):
    table = get_monte_carlo(name)
    oilcane_price = table[oc.MFPP.index].to_numpy()[:, np.newaxis] # USD/ton
    biodiesel_flow = table[oc.biodiesel_production.index].to_numpy()[:, np.newaxis] * 1e6 # gal/yr
    ethanol_price_baseline = table[oc.set_ethanol_price.index].to_numpy()[:, np.newaxis]
    biodiesel_price_baseline = table[oc.set_biodiesel_price.index].to_numpy()[:, np.newaxis]
    ethanol_flow = table[oc.ethanol_production.index].to_numpy()[:, np.newaxis] * 1e6 # gal/yr
    feedstock_flow = table[oc.feedstock_consumption.index].to_numpy()[:, np.newaxis] # ton/yr
    baseline_price = (
        oilcane_price
        - (ethanol_price_baseline * ethanol_flow + biodiesel_price_baseline * biodiesel_flow) / feedstock_flow
    )
    return (
        baseline_price 
        + (ethanol_price[np.newaxis, :] * ethanol_flow + biodiesel_price[np.newaxis, :] * biodiesel_flow) / feedstock_flow
    )

def evaluate_MFPP_benefit_uncertainty_across_ethanol_and_biodiesel_prices(name, ethanol_price, biodiesel_price, baseline=None):
    if baseline is None:
        configuration = parse(name)
        number, agile = configuration
        assert number > 0
        baseline = Configuration(-number, agile)
    MFPP_baseline = evaluate_MFPP_uncertainty_across_ethanol_and_biodiesel_prices(baseline, ethanol_price, biodiesel_price)
    MFPP = evaluate_MFPP_uncertainty_across_ethanol_and_biodiesel_prices(name, ethanol_price, biodiesel_price)
    return MFPP - MFPP_baseline
@no_derivative
def evaluate_MFPP_across_ethanol_and_biodiesel_prices(ethanol_price, biodiesel_price, configuration=None):
    if configuration is not None: oc.load(configuration)
    feedstock_flow = oc.flows['feedstock']()
    biodiesel_flow = oc.flows['biodiesel']()
    ethanol_flow = oc.flows['ethanol']()
    baseline_price = (
        oc.tea.solve_price(oc.oilcane) * oc.kg_per_ton
        - (oc.ethanol.price * ethanol_flow  * 2.98668849 + oc.biodiesel.price * 3.3111 * biodiesel_flow) / feedstock_flow
    )
    return (
        baseline_price 
        + (ethanol_price * ethanol_flow + biodiesel_price * biodiesel_flow) / feedstock_flow
    )

def evaluate_MFPP_benefit_across_ethanol_and_biodiesel_prices(ethanol_price, biodiesel_price, baseline=None, configuration=None):
    if configuration is None: configuration = oc.configuration
    if baseline is None:
        number, agile = configuration
        assert number > 0
        baseline = Configuration(-number, agile)
    MFPP_baseline = evaluate_MFPP_across_ethanol_and_biodiesel_prices(ethanol_price, biodiesel_price, baseline)
    MFPP = evaluate_MFPP_across_ethanol_and_biodiesel_prices(ethanol_price, biodiesel_price, configuration)
    return MFPP - MFPP_baseline

def run_uncertainty_and_sensitivity(name, N, rule='L',
                                    across_oil_content=False, 
                                    sample_cache={},
                                    autosave=True,
                                    autoload=True):
    np.random.seed(1)
    from warnings import filterwarnings
    filterwarnings('ignore', category=bst.utils.DesignWarning)
    oc.load(name, cache=None)
    key = (N, rule)
    if key in sample_cache:
        samples = sample_cache[key]
    else:
        sample_cache[key] = samples = oc.model.sample(N, rule)
    oc.model.load_samples(samples)
    file = monte_carlo_file(name, across_oil_content)
    if across_oil_content:
        if parse(name).number < 0:
            oc.model.evaluate(notify=int(N/10))
            oc.model.evaluate_across_coordinate(
                name='Oil content',
                f_coordinate=lambda x: None,
                coordinate=oc.oil_content,
                notify=int(N/10), 
                notify_coordinate=True,
                xlfile=file,
                re_evaluate=False,
            )
        else:
            def f(x):
                oc.oil_extraction_specification.locked_oil_content = False
                oc.oil_extraction_specification.load_oil_content(x)
                oc.oil_extraction_specification.locked_oil_content = True
            oc.model.evaluate_across_coordinate(
                name='Oil content',
                f_coordinate=f,
                coordinate=oc.oil_content,
                notify=int(N/10), 
                notify_coordinate=True,
                xlfile=file,
            )
    else:
        N = min(int(N/10), 20)
        autosave = N if autosave else False
        autoload_file = autoload_file_name(name)
        success = False
        for i in range(3):
            try:
                oc.model.evaluate(
                    notify=N,
                    autosave=autosave,
                    autoload=autoload,
                    file=autoload_file
                )
            except Exception as e:
                raise e from None
                warn('failed evaluation; restarting without cache')
            else:
                success = True
                break
        if not success:
            raise RuntimeError('evaluation failed')
        oc.model.table.to_excel(file)
        rho, p = oc.model.spearman_r()
        file = spearman_file(name)
        rho.to_excel(file)

run = run_uncertainty_and_sensitivity
    
def run_all(N, across_oil_content=False, rule='L', configurations=None, **kwargs):
    if configurations is None: configurations = oc.configuration_names
    for name in configurations:
        print(f"Running {name}:")
        run_uncertainty_and_sensitivity(
            name, N, rule, across_oil_content, **kwargs
        )


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import biosteam as bst
import pandas as pd

__all__ = ('ConventionalEthanolTEA', 'create_tea', 'capex_table', 'foc_table')

class ConventionalEthanolTEA(bst.TEA):
    """
    Create a ConventionalEthanolTEA object for techno-economic analysis of a
    biorefinery [1]_.
    
    Parameters
    ----------    
    system : System
        Should contain feed and product streams.
    IRR : float
        Internal rate of return (fraction).
    duration : tuple[int, int]
        Start and end year of venture (e.g. (2018, 2038)).
    depreciation : str
        'MACRS' + number of years (e.g. 'MACRS7').
    operating_days : float
        Number of operating days per year.
    income_tax : float
        Combined federal and state income tax rate (fraction).
    lang_factor : float
        Lang factor for getting fixed capital investment from
        total purchase cost. If no lang factor, estimate capital investment
        using bare module factors.
    startup_schedule : tuple[float]
        Startup investment fractions per year 
        (e.g. (0.5, 0.5) for 50% capital investment in the first year and 50%
        investment in the second).
    WC_over_FCI : float
        Working capital as a fraction of fixed capital investment.
    labor_cost : float
        Total labor cost (USD/yr).
    fringe_benefits : float
        Cost of fringe benefits as a fraction of labor cost.
    property_tax : float
        Fee as a fraction of fixed capital investment.
    property_insurance : float
        Fee as a fraction of fixed capital investment.    
    supplies : float
        Yearly fee as a fraction of labor cost.
    maintenance : float
        Yearly fee as a fraction of fixed capital investment.
    administration : float
        Yearly fee as a fraction of fixed capital investment.

    References
    ----------
    .. [1] Huang, H., Long, S., & Singh, V. (2016). Techno-economic analysis of biodiesel
        and ethanol co-production from lipid-producing sugarcane. Biofuels, Bioproducts
        and Biorefining, 10(3), 299–315. https://doi.org/10.1002/bbb.1640
    
    """
    __slots__ = ('labor_cost', 'fringe_benefits', 'maintenance',
                 'property_tax', 'property_insurance', '_FCI_cached',
                 'supplies', 'maintanance', 'administration')
    
    def __init__(self, system, IRR, duration, depreciation, income_tax,
                 operating_days, lang_factor, construction_schedule, WC_over_FCI,
                 labor_cost, fringe_benefits, property_tax,
                 property_insurance, supplies, maintenance, administration):
        super().__init__(system, IRR, duration, depreciation, income_tax,
                         operating_days, lang_factor, construction_schedule,
                         startup_months=0, startup_FOCfrac=0, startup_VOCfrac=0,
                         startup_salesfrac=0, finance_interest=0, finance_years=0, 
                         finance_fraction=0, WC_over_FCI=WC_over_FCI)
        self.labor_cost = labor_cost
        self.fringe_benefits = fringe_benefits
        self.property_tax = property_tax
        self.property_insurance = property_insurance
        self.supplies= supplies
        self.maintenance = maintenance
        self.administration = administration
    
    def _FCI(self, TDC):
        self._FCI_cached = TDC
        return TDC
    
    def _FOC(self, FCI):
        return (FCI*(self.property_tax + self.property_insurance
                     + self.maintenance + self.administration)
                + self.labor_cost*(1+self.fringe_benefits+self.supplies))
    
def create_tea(system, cls=ConventionalEthanolTEA):
    return cls(system, IRR=0.15,
               duration=(2018, 2038),
               depreciation='MACRS7', income_tax=0.35,
               operating_days=200, lang_factor=3,
               construction_schedule=(0.4, 0.6), WC_over_FCI=0.05,
               labor_cost=2.5e6, fringe_benefits=0.4,
               property_tax=0.001, property_insurance=0.005,
               supplies=0.20, maintenance=0.01, administration=0.005)

def capex_table(tea):
    purchase_cost = tea.purchase_cost /1e6
    lang_factor = tea.lang_factor
    FCI = lang_factor * purchase_cost
    working_capital = FCI * tea.WC_over_FCI
    TCI = FCI + working_capital
    index = ('Purchase cost',
             'Fixed capital investment', 
             'Working capital', 
             'Total capital investment')
    data = [
        ['', purchase_cost],
        [f"Purchase cost x {lang_factor}", FCI],
        [f"{tea.WC_over_FCI:.1%} of FCI", working_capital],
        ['', TCI],
    ]
    return pd.DataFrame(data, index=index, columns=['Notes', 'Cost [MM$]'])

def foc_table(tea):
    foc = bst.report.FOCTableBuilder()
    FCI = tea.FCI / 1e6
    labor_cost = tea.labor_cost / 1e6
    labor_burden = tea.fringe_benefits + tea.supplies
    foc.entry('Labor salary', labor_cost)
    foc.entry('Labor burden', labor_cost * labor_burden, f'{labor_burden:.0%} of labor salary')
    foc.entry('Maintenance', tea.maintenance * FCI, f'{tea.maintenance:.0%} of FCI')
    foc.entry('Administration', tea.administration * FCI, f'{tea.administration:.1%} of FCI')
    foc.entry('Property tax', tea.property_tax * FCI, f'{tea.property_tax:.1%} of FCI')
    foc.entry('Property insurance', tea.property_insurance * FCI, f'{tea.property_insurance:.1%} of FCI')
    return foc.table()


# -*- coding: utf-8 -*-
"""
Created on Thu Nov  4 14:44:17 2021

@author: yrc2
"""
from biosteam import MockVariable

(set_cane_oil_content, set_relative_sorghum_oil_content, set_bagasse_oil_retention, 
 set_bagasse_oil_extraction_efficiency, 
 set_cane_operating_days, set_sorghum_operating_days, 
 set_plant_capacity, set_ethanol_price,
 set_biodiesel_price, set_natural_gas_price, 
 set_electricity_price, 
 set_IRR, set_crude_glycerol_price, set_pure_glycerol_price,
 set_saccharification_reaction_time, set_cellulase_price, 
 set_cellulase_loading, set_reactor_base_cost,
 set_cane_glucose_yield, set_cane_xylose_yield, 
 set_sorghum_glucose_yield, set_sorghum_xylose_yield, 
 set_glucose_to_ethanol_yield, set_xylose_to_ethanol_yield,
 set_cofermentation_titer, set_cofermentation_productivity,
 set_cane_PL_content, set_sorghum_PL_content, set_cane_FFA_content,
 set_sorghum_FFA_content, set_TAG_to_FFA_conversion, set_oilcane_GWP,
 set_methanol_GWP, set_pure_glycerine_GWP, set_cellulase_GWP, set_natural_gas_GWP
 ) = all_parameter_mockups = (
    MockVariable('Oil retention', '%', 'biorefinery'),
    MockVariable('Bagasse oil extraction efficiency', '%', 'biorefinery'),
    MockVariable('Cane operating days', 'day/yr', 'biorefinery'),
    MockVariable('Sorghum operating days', 'day/yr', 'biorefinery'),
    MockVariable('Annual crushing capacity', 'ton/yr', 'biorefinery'),
    MockVariable('Price', 'USD/gal', 'Stream-ethanol'),
    MockVariable('Price', 'USD/gal', 'Stream-biodiesel'),
    MockVariable('Price', 'USD/cf', 'Stream-natural gas'),
    MockVariable('Electricity price', 'USD/kWhr', 'biorefinery'),
    MockVariable('IRR', '%', 'biorefinery'),
    MockVariable('Price', 'USD/kg', 'Stream-crude glycerol'),
    MockVariable('Price', 'USD/kg', 'Stream-pure glycerine'),
    MockVariable('Reaction time', 'hr', 'Saccharification'),
    MockVariable('Price', 'USD/kg', 'cellulase'),
    MockVariable('Cellulase loading', 'wt. % cellulose', 'cellulase'),
    MockVariable('Base cost', 'million USD', 'Pretreatment reactor system'),
    MockVariable('Cane glucose yield', '%', 'Pretreatment and saccharification'),
    MockVariable('Sorghum glucose yield', '%', 'Pretreatment and saccharification'),
    MockVariable('Cane xylose yield', '%', 'Pretreatment and saccharification'),
    MockVariable('Sorghum xylose yield', '%', 'Pretreatment and saccharification'),
    MockVariable('Glucose to ethanol yield', '%', 'Cofermenation'),
    MockVariable('Xylose to ethanol yield', '%', 'Cofermenation'),
    MockVariable('Titer', 'g/L', 'Cofermentation'),
    MockVariable('Productivity', 'g/L', 'Cofermentation'),
    MockVariable('Cane PL content', '% oil', 'oilcane'),
    MockVariable('Sorghum PL content', '% oil', 'oilsorghum'),
    MockVariable('Cane FFA content', '% oil', 'oilcane'),
    MockVariable('Sorghum FFA content', '% oil', 'oilsorghum'),
    MockVariable('Cane oil content', 'dry wt. %', 'oilcane'),
    MockVariable('Relative sorghum oil content', 'dry wt. %', 'oilsorghum'),
    MockVariable('TAG to FFA conversion', '% oil', 'biorefinery'),
    MockVariable('GWP', 'kg*CO2-eq/kg', 'Stream-oilcane'),
    MockVariable('GWP', 'kg*CO2-eq/kg', 'Stream-methanol'),
    MockVariable('GWP', 'kg*CO2-eq/kg', 'Stream-pure glycerine'),
    MockVariable('GWP', 'kg*CO2-eq/kg', 'Stream-cellulase'),
    MockVariable('GWP', 'kg*CO2-eq/kg', 'Stream-natural gas'),
)
     
(MFPP, feedstock_consumption, biodiesel_production, ethanol_production, 
 electricity_production, natural_gas_consumption, TCI, 
 heat_exchanger_network_error, GWP_economic, GWP_ethanol, GWP_biodiesel, 
 GWP_crude_glycerol, GWP_electricity, GWP_ethanol_displacement,
 GWP_biofuel_allocation, GWP_ethanol_allocation,
 GWP_biodiesel_allocation, GWP_crude_glycerol_allocation,
 MFPP_derivative, 
 biodiesel_production_derivative, ethanol_production_derivative, 
 electricity_production_derivative, natural_gas_consumption_derivative, 
 TCI_derivative, GWP_economic_derivative, 
 GWP_ethanol_derivative, GWP_biodiesel_derivative,
 GWP_crude_glycerol_derivative, GWP_electricity_derivative
 ) = all_metric_mockups = (
    MockVariable('MFPP', 'USD/ton', 'Biorefinery'),
    MockVariable('Feedstock consumption', 'ton/yr', 'Biorefinery'),
    MockVariable('Biodiesel production', 'Gal/ton', 'Biorefinery'),
    MockVariable('Ethanol production', 'Gal/ton', 'Biorefinery'),
    MockVariable('Electricity production', 'kWhr/ton', 'Biorefinery'),
    MockVariable('Natural gas consumption', 'cf/ton', 'Biorefinery'),
    MockVariable('TCI', '10^6*USD', 'Biorefinery'),
    MockVariable('Heat exchanger network error', '%', 'Biorefinery'),
    MockVariable('GWP', 'kg*CO2*eq / USD', 'Economic allocation'),
    MockVariable('Ethanol GWP', 'kg*CO2*eq / gal', 'Economic allocation'),
    MockVariable('Biodiesel GWP', 'kg*CO2*eq / gal', 'Economic allocation'),
    MockVariable('Crude glycerol GWP', 'kg*CO2*eq / kg', 'Economic allocation'),
    MockVariable('Electricity GWP', 'kg*CO2*eq / MWhr', 'Economic allocation'),
    MockVariable('Ethanol GWP', 'kg*CO2*eq / gal', 'Displacement allocation'),
    MockVariable('Biofuel GWP', 'kg*CO2*eq / GGE', 'Energy allocation'),
    MockVariable('Ethanol GWP', 'kg*CO2*eq / gal', 'Energy allocation'),
    MockVariable('Biodiesel GWP', 'kg*CO2*eq / gal', 'Energy allocation'),
    MockVariable('Crude-glycerol GWP', 'kg*CO2*eq / kg', 'Energy allocation'),
    MockVariable('MFPP derivative', 'USD/ton', 'Biorefinery'),
    MockVariable('Biodiesel production derivative', 'Gal/ton', 'Biorefinery'),
    MockVariable('Ethanol production derivative', 'Gal/ton', 'Biorefinery'),
    MockVariable('Electricity production derivative', 'kWhr/ton', 'Biorefinery'),
    MockVariable('Natural gas consumption derivative', 'cf/ton', 'Biorefinery'),
    MockVariable('TCI derivative', '10^6*USD', 'Biorefinery'),
    MockVariable('GWP derivative', 'kg*CO2*eq / USD', 'Economic allocation'),
    MockVariable('Ethanol GWP derivative', 'kg*CO2*eq / gal', 'Ethanol'),
    MockVariable('Biodiesel GWP derivative', 'kg*CO2*eq / gal', 'Biodiesel'),
    MockVariable('Crude glycerol GWP derivative', 'kg*CO2*eq / kg', 'Crude glycerol'),
    MockVariable('Electricity GWP derivative', 'kg*CO2*eq / MWhr', 'Electricity'),
)

tea_monte_carlo_metric_mockups = (
    MFPP, 
    TCI,
    ethanol_production,
    biodiesel_production,
    electricity_production,
    natural_gas_consumption
)

tea_monte_carlo_derivative_metric_mockups = (
    MFPP_derivative, 
    TCI_derivative,
    ethanol_production_derivative,
    biodiesel_production_derivative,
    electricity_production_derivative,
    natural_gas_consumption_derivative,
)

lca_monte_carlo_metric_mockups = (
    GWP_economic,
    GWP_ethanol,
    GWP_biodiesel,
    GWP_electricity,
    GWP_crude_glycerol,
)

lca_monte_carlo_derivative_metric_mockups = (
    GWP_economic_derivative,
    GWP_ethanol_derivative,
    GWP_biodiesel_derivative, 
    GWP_electricity_derivative,
    GWP_crude_glycerol_derivative,
)